;HD44780-compatible VFD driver for 16 character alphanumeric VFDs (14 Segments)
;ATTiny2313 @ 8 MHz internal R/C osc.
;
;PORTS:
;  PA0:   shift register "Clock"
;  PA1:   shift register "Data"
;  PB0-7: Data bus (HD44780 compatible)
;  PD0:   shift register "Strobe"
;  PD1:   HD44780 "RS"
;  PD2:   HD44780 "R/W'"
;  PD3:   HD44780 "E"
;  PD4-6: address lines A0, A1, A2 for VFD grid selection
;SHIFT REGISTER:
;  16 stage, Q0..Q15:
;    Q15-Q8: VFD segments a, b, c, d, e, f, g1, g2
;    Q7-Q1:  VFD segments h, i, j, k, l, m, dp
;    Q0:     address line A3 for VFD grid selection
;VFD SEGMENTS:
;          a
;    ===========
;   | \   |   / |
; f |  \h |i /j | b
;   |   \ | /   |
;   | g1 \|/ g2 |
;    ===== =====
;   |    /|\    |
; e |   / | \   | c
;   | k/ l| m\  |
;   | /   |   \ |   _
;    ===========   [_]
;       d          dp

;CHANGELOG:
;  v0.2  2006-05-21  Arne Rossius
;  v0.3  2006-10-23  Arne Rossius
;    * display version info on power-up

.include "../include/tn2313def.inc"

.equ	Clock = 0
.equ	Data = 1
.equ	Strobe = 0
.equ	RS = 1
.equ	RW = 2
.equ	E = 3

.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	grid = R19
.def	disp1 = R20
.def	disp2 = R21
.def	nibble = R22
.def	ddpos = R23
.def	cgpos = R24
.def	shift = R25

.def	flags = R27
	.equ	AutoInc = 0 ;auto inc (1) or dec (0)
	.equ	AutoShift = 1 ;auto display shift
	.equ	DispOn = 2 ;display on
	.equ	TwoLines = 3 ;two display lines (emulate 2x8 instead of 1x16)
	.equ	Interface = 4 ;interface (0 = 8 bits, 1 = 4 bits)
	.equ	NibbleRx = 5 ;one nibble received
	.equ	CGRAM = 6 ;data is for CGRAM

;R28/29 = Y pointer
;R30/31 = Z pointer

;===============================================================================

.dseg
RAM_DD:		.byte 80 ;80 chars, ASCII code
RAM_CG:		.byte 16 ;8 user defined chars, 2 bytes each
RAM_STACK:	.byte 32

;===============================================================================

.cseg
.org 0x000
	rjmp	start
.org INT0addr
	rjmp	int_rw
.org INT1addr
	rjmp	int_e

;===============================================================================

int_rw:
	;TODO
	reti

;--------------------

int_e: ;falling edge on "E" pin
	push	temp
	push	temp2
	push	count
	push	YL
	in	temp, SREG
	push	temp
	
	in	temp, PINB
	sbrs	flags, Interface
	rjmp	int_e_complete
	;4 bit mode:
	sbrs	flags, NibbleRx
	clr	nibble
	swap	nibble
	swap	temp
	andi	temp, 0x0F
	add	nibble, temp
	mov	temp, nibble
	sbrc	flags, NibbleRx
	rjmp	int_e_complete
	ori	flags, 1<<NibbleRx
	rjmp	int_e_end

int_e_complete:
	;data or command?
	andi	flags, ~(1<<NibbleRx)
	sbic	PIND, RS
	rjmp	int_e_data
	
	;HD44780 command:
	sbrc	temp, 7
	rjmp	ddram_addr
	sbrc	temp, 6
	rjmp	cgram_addr
	sbrc	temp, 5
	rjmp	function_set
	sbrc	temp, 4
	rjmp	display_shift
	sbrc	temp, 3
	rjmp	display_on_off
	sbrc	temp, 2
	rjmp	entry_mode
	sbrc	temp, 1
	rjmp	return_home
	sbrc	temp, 0
	rjmp	clear_disp
	rjmp	int_e_end

clear_disp: ;clear display and set DDRAM address to 0x00
	ldi	temp, ' '
	ldi	count, 80
	ldi	YL, RAM_DD
clear_disp_loop:
	st	Y+, temp
	dec	count
	brne	clear_disp_loop
	ldi	ddpos, 0x00
	rjmp	int_e_end

return_home: ;remove display shifting and set DDRAM address to 0x00
	ldi	shift, 0
	ldi	ddpos, 0x00
	rjmp	int_e_end

entry_mode: ;set AutoIncrement/Decrement and AutoShift
	andi	flags, ~(1<<AutoInc)
	sbrc	temp, 1
	ori	flags, 1<<AutoInc
	andi	flags, ~(1<<AutoShift)
	sbrc	temp, 0
	ori	flags, ~(1<<AutoShift)
	rjmp	int_e_end

display_on_off: ;set Display On/Off and Cursor
	andi	flags, ~(1<<DispOn)
	sbrc	temp, 2
	ori	flags, 1<<DispOn
	;TODO cursor
	rjmp	int_e_end

display_shift: ;shift display and/or move cursor
	;TODO
	rjmp	int_e_end

function_set: ;set Interface Mode (4/8 bit) and number of lines (1 or 2)
	andi	flags, ~(1<<Interface)
	sbrs	temp, 4
	ori	flags, 1<<Interface
	andi	flags, ~(1<<TwoLines)
	sbrc	temp, 3
	ori	flags, 1<<TwoLines
	andi	flags, ~(1<<NibbleRx)
	;(ignore font setting)
	rjmp	int_e_end

cgram_addr: ;set CGRAM Address
	andi	temp, 0x3F
	mov	cgpos, temp
	ori	flags, 1<<CGRAM
	rjmp	int_e_end

ddram_addr: ;set DDRAM Address
	andi	flags, ~(1<<CGRAM)
	andi	temp, 0x7F
	sbrc	flags, TwoLines
	rjmp	ddram_twolines
	cpi	temp, 80
	brlo	PC+2
	subi	temp, 80
	mov	ddpos, temp
	rjmp	int_e_end
ddram_twolines:
	cpi	temp, 0x40
	brsh	ddram_line2
	cpi	temp, 40
	brlo	PC+2
	subi	temp, 40
	mov	ddpos, temp
	rjmp	int_e_end
ddram_line2:
	subi	temp, 0x40-40
	cpi	temp, 80
	brlo	PC+2
	subi	temp, 40
	mov	ddpos, temp
	rjmp	int_e_end

int_e_data:
	;HD44780 data byte
	sbrc	flags, CGRAM
	rjmp	int_e_cgdata
	ldi	YL, RAM_DD
	add	YL, ddpos
	st	Y, temp
	;inc/dec address
	inc	ddpos
	sbrs	flags, AutoInc
	subi	ddpos, -2
	cpi	ddpos, 80 ;inc wrap
	brne	PC+2
	ldi	ddpos, 0
	cpi	ddpos, 0xFF ;dec wrap
	brne	PC+2
	ldi	ddpos, 79
	rjmp	int_e_end
	
int_e_cgdata:
	;HD44780 CGRAM data (4 bytes @ 4 bits [lower, high nibble first] + 4 bytes [ignored])
	sbrc	cgpos, 2 ;ignore 2nd half of data
	rjmp	int_e_cgdata_end
	ldi	YL, 0x00 ;convert to RAM address
	sbrc	cgpos, 1
	ori	YL, 0x01
	sbrc	cgpos, 3
	ori	YL, 0x02
	sbrc	cgpos, 4
	ori	YL, 0x04
	sbrc	cgpos, 5
	ori	YL, 0x08
	subi	YL, -RAM_CG
	sbrs	cgpos, 0 ;lower nibble?
	swap	temp
	andi	temp, 0x0F
	ld	temp2, Y
	andi	temp2, 0xF0
	or	temp, temp2
	st	Y, temp ;store new data
int_e_cgdata_end:
	inc	cgpos ;auto-increment (TODO: decrement if AutoInc = 0?)
	cpi	cgpos, 64
	brlo	PC+2
	ldi	cgpos, 0
	rjmp	int_e_end

int_e_end:
	;return
	pop	temp
	out	SREG, temp
	pop	YL
	pop	count
	pop	temp2
	pop	temp
	reti

;===============================================================================

initmsg:
	.db "LCD2VFD V0.3 RDY"

start:
	;set stackpointer
	ldi	temp, LOW(RAMEND)
	out	SPL, temp

	;init ports
	ldi	temp, 0x03
	out	DDRA, temp
	ldi	temp, 0x00
	out	PORTA, temp
	
	ldi	temp, 0x00
	out	DDRB, temp
	ldi	temp, 0xFF
	out	PORTB, temp

	ldi	temp, 0x71
	out	DDRD, temp
	ldi	temp, 0x0E
	out	PORTD, temp

	;enable external interrupts
	ldi	temp, 0x09 ;INT0 (R/W): any edge, INT1 (E): falling edge
	out	MCUCR, temp
	ldi	temp, 1<<INT1 ; | 1<<INT0
	out	GIMSK, temp

	;init registers
	ldi	YH, 0
	ldi	flags, 1<<DispOn
	ldi	ddpos, 0
	ldi	cgpos, 0
	ldi	shift, 0

	;fill display with spaces
	ldi	count, 16
	ldi	YL, RAM_DD
	ldi	ZL, LOW(initmsg*2)
	ldi	ZH, HIGH(initmsg*2)
initram:
	lpm	temp, Z+
	st	Y+, temp
	dec	count
	brne	initram
	ldi	count, 80-16
	ldi	temp, ' '
initram2:
	st	Y+, temp
	dec	count
	brne	initram2

	;enable interrupts
	sei

;===============================================================================

	ldi	grid, 0
mux:
	;get char
	mov	YL, shift
	add	YL, grid
	cpi	YL, 80
	brlo	PC+2
	subi	YL, 80
	subi	YL, -RAM_DD
	sbrs	flags, TwoLines
	rjmp	mux_load
	sbrs	grid, 3 ;grid >= 8 (2nd display half)?
	rjmp	mux_load
	subi	YL, -32
mux_load:
	ld	temp, Y+
	sbrs	flags, DispOn
	ldi	temp, ' ' ;display spaces if display is "off"
	;get values for 14 segment display
	cpi	temp, 16
	brlo	mux_custom ;custom char
	ldi	ZL, LOW(charset*2)
	ldi	ZH, HIGH(charset*2)
	add	ZL, temp
	brcc	PC+2
	inc	ZH
	add	ZL, temp
	brcc	PC+2
	inc	ZH
	lpm	disp1, Z+
	lpm	disp2, Z+
	rjmp	mux_disp
	
mux_custom:
	;get custom char from RAM
	andi	temp, 0x07
	lsl	temp
	ldi	YL, RAM_CG
	add	YL, temp
	ld	disp1, Y+
	ld	disp2, Y+
	
mux_disp:
	;display char
	rcall	shiftout
	mov	temp, grid
	swap	temp
	andi	temp, 0x70
	ori	temp, 1<<E | 1<<RW | 1<<RS | 1<<Strobe
	ori	temp, 0x0E ;pullups
	out	PORTD, temp
	nop
	cbi	PORTD, Strobe
	;delay
	ldi	temp, 5
	ldi	temp2, 0
delay:
	dec	temp2
	brne	delay
	dec	temp
	brne	delay
	;next grid
	inc	grid
	cpi	grid, 16
	brlo	mux
	ldi	grid, 0
	rjmp	mux

;--------------------

shiftout: ;invert and shift out disp1[0..7], disp2[0..6] and grid[3]
	mov	temp, disp1
	ldi	temp2, 8
shiftout1:
	sbi	PORTA, Data
	lsr	temp
	brcc	PC+2
	cbi	PORTA, Data
	sbi	PORTA, Clock
	nop
	cbi	PORTA, Clock
	dec	temp2
	brne	shiftout1
	
	mov	temp, disp2
	ldi	temp2, 7
shiftout2:
	sbi	PORTA, Data
	lsr	temp
	brcc	PC+2
	cbi	PORTA, Data
	sbi	PORTA, Clock
	nop
	cbi	PORTA, Clock
	dec	temp2
	brne	shiftout2

	cbi	PORTA, Data
	sbrc	grid, 3
	sbi	PORTA, Data
	sbi	PORTA, Clock
	nop
	cbi	PORTA, Clock

	ret

;===============================================================================

charset:
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 ;0x00
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 ;0x10
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
	.dw 0x0000, 0x0709, 0x0220, 0x12ce, 0x12ed, 0x0c24, 0x24f9, 0x0200 ;0x20
	.dw 0x2400, 0x0900, 0x3fc0, 0x12c0, 0x0800, 0x00c0, 0x4000, 0x0c00
	.dw 0x0c3f, 0x0406, 0x088b, 0x048d, 0x00e6, 0x2069, 0x00fd, 0x0c01 ;0x30
	.dw 0x00ff, 0x00ef, 0x0009, 0x0801, 0x2400, 0x00c8, 0x0900, 0x1083
	.dw 0x08af, 0x00f7, 0x128f, 0x0039, 0x120f, 0x0079, 0x0071, 0x00bd ;0x40
	.dw 0x00f6, 0x1209, 0x001e, 0x2470, 0x0038, 0x0536, 0x2136, 0x003f
	.dw 0x00f3, 0x203f, 0x20f3, 0x00ed, 0x1201, 0x003e, 0x0c30, 0x2836 ;0x50
	.dw 0x2d00, 0x0d00, 0x0c09, 0x0039, 0x2100, 0x000f, 0x0001, 0x0008
	.dw 0x0100, 0x00f7, 0x128f, 0x0039, 0x120f, 0x0079, 0x0071, 0x00bd ;0x60
	.dw 0x00f6, 0x1209, 0x001e, 0x2470, 0x0038, 0x0536, 0x2136, 0x003f
	.dw 0x00f3, 0x203f, 0x20f3, 0x00ed, 0x1201, 0x003e, 0x0c30, 0x2836 ;0x70
	.dw 0x2d00, 0x0d00, 0x0c09, 0x2440, 0x1200, 0x0980, 0x0001, 0x2808
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 ;0x80
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 ;0x90
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 ;0xA0
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 ;0xB0
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 ;0xC0
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 ;0xD0
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 ;0xE0
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 ;0xF0
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000

;===============================================================================

;TODO/DEBUG:
hex:
	push	temp
	swap	temp
	andi	temp, 0x0F
	ori	temp, 0x30
	cpi	temp, 58
	brlo	PC+2
	subi	temp, -7
	st	Y+, temp
	pop	temp
	andi	temp, 0x0F
	ori	temp, 0x30
	cpi	temp, 58
	brlo	PC+2
	subi	temp, -7
	st	Y+, temp
	ret
	