;HD44780-compatible VFD driver for 16 character alphanumeric VFDs (14 Segments)
;ATTiny2313 @ 8 MHz internal R/C osc.
;
;Receive a text via RS232 (9600 Baud) and scroll it indefinitely
;Clear buffer for new text: send character #12 (Form Feed)

.equ	TextBuffer = 96 ;size of text buffer

;PORTS:
;  PA0:   shift register "Clock"
;  PA1:   shift register "Data"
;  PB0-7: [unused]
;  PD0:   shift register "Strobe"
;  PD1:   [unused]
;  PD2:   RS232 Rx input
;  PD3:   always on (power for external RS232 conversion circuitry)
;  PD4-6: binary-to-hex-decoder address lines A0, A1, A2 for VFD grid selection
;
;SHIFT REGISTER:
;  16 stage, Q0..Q15:
;    Q15-Q8: VFD segments a, b, c, d, e, f, g1, g2
;    Q7-Q1:  VFD segments h, i, j, k, l, m, dp
;    Q0:     binary-to-hex-decoder address line A3 for VFD grid selection
;
;VFD SEGMENTS:
;          a
;    ===========
;   | \   |   / |
; f |  \h |i /j | b
;   |   \ | /   |
;   | g1 \|/ g2 |
;    ===== =====
;   |    /|\    |
; e |   / | \   | c
;   | k/ l| m\  |
;   | /   |   \ |   _
;    ===========   [_]
;       d          dp
;
;CHANGELOG:
;  v0.1  2006-05-20  Arne Rossius
;  v0.5  2006-10-23  Arne Rossius

.include "../include/tn2313def.inc"

.equ	Clock = 0
.equ	Data = 1
.equ	Strobe = 0
.equ	RS = 1
.equ	RW = 2
.equ	E = 3

;Software UART: (9600 Baud @ 8 MHz)
.equ	u_N = 104 ;Timer Ticks per Bit
.equ	u_C = 8 ;Divisor
.equ	u_R = 0x02 ;Timer0 Divisor Select (CLK/8)
.def	u_buffer = R14 ;Serial buffer
.def	u_sr = R15 ;Status-register storage
.def	u_tmp = R16 ;Scratchregister
.def	u_bit_cnt = R17 ;Bit counter
.def	u_reload = R18 ;Reload-register (internal - do not use)
.def	u_transmit = R19 ;Data to transmit

.def	flags = R20
	.equ	RDR = 0 ;new byte received
	.equ	Scroll = 1 ;scroll text (length > 16)
.def	temp = R21
.def	temp2 = R22
.def	count = R23
.def	grid = R24
.def	disp1 = R25
.def	disp2 = R26
.def	buffer = R27
;R28/29 = Y pointer (display RAM)
;R30/31 = Z pointer (text RAM = RS232 Buffer)

;===============================================================================

.dseg
RAM_Disp:	.byte 16
RAM_Buffer:	.byte TextBuffer
RAM_BufferEnd:
RAM_Stack:	.byte 16

;===============================================================================

.cseg
.org 0x000
	rjmp	start
.org INT0addr
	rjmp	ext_int0 ;External interrupt handler
.org OVF0addr
	rjmp	tim0_ovf ;Timer0 overflow handler

;===============================================================================

;Software UART:
ext_int0:
	in	u_sr,SREG		;Store Status Register
	
	ldi	u_tmp,(256-(u_N+u_N/2)+(29/u_C));Set timer reload-value (to 1.5
	out	TCNT0,u_tmp		;  bit len). 29 = time delay that
					;  have already been used in this
					;  interrupt plus the time
					;  that will be used by the time
					;  delay between timer interrupt request
					;  and the actual sampling of the first
					;  data bit.

	ldi	u_tmp,1<<TOIE0		;Set bit 1 in u_tmp
	out	TIFR,u_tmp		;   to clear T/C0 overflow flag
	out	TIMSK,u_tmp		;   and enable T/C0 overflow interrupt

	clr	u_bit_cnt		;Clear bit counter
	out	GIMSK,u_bit_cnt		;Disable external interrupt

	ldi	u_reload,(256-u_N+(8/u_C));Set reload-value (constant).

	out	SREG,u_sr		;Restore SREG
	reti

;===============================================================================

;Software UART:
tim0_ovf:
	in	u_sr,SREG		;Store statusregister

	out	TCNT0,u_reload		;Reload timer

	inc	u_bit_cnt		;Increment bit_counter
	
	;receive:
	sec				;Set carry
	sbis	PIND,2			;if PD2=LOW         <=== SAMPLE HERE
	clc				;    clear carry
	ror	u_buffer		;Shift carry into data
	in	u_tmp,SREG		;Store SREG

	cpi	u_bit_cnt,9		;if u_bit_cnt!=9 (must sample stop-bit)
	brne	tim0_ret		;   exit interrupt

	out	SREG,u_tmp		;Get old SREG
	rol	u_buffer		;Rotate back data (to get rid of the stop-bit)

	sbr	flags, 1<<RDR

	;complete:
	ldi	u_tmp,1<<INTF0
	out	GIFR,u_tmp		;clear pending ext. int.
	ldi	u_tmp,1<<INT0		;(u_bit_cnt==9):
	out	GIMSK,u_tmp		;Enable external interrupt
	clr	u_tmp 
	out	TIMSK,u_tmp		;Disable timer interrupt
	mov	u_tmp, u_buffer
	cpi	u_tmp, 32 ;save if >= 32 (no control char)
	brsh	save_char

	;control chars:
	cpi	u_tmp, 8 ;Backspace -> remove last char from buffer
	brne	backspace_end
	cpi	buffer, RAM_Buffer
	breq	backspace_end
	dec	buffer
backspace_end:
	cpi	u_tmp, 10 ;Line Feed -> clear buffer
	brne	linefeed_end
	ldi	buffer, RAM_Buffer
linefeed_end:
	cpi	u_tmp, 12 ;Form Feed -> clear buffer
	brne	formfeed_end
	ldi	buffer, RAM_Buffer
formfeed_end:
	cpi	u_tmp, 13 ;Carriage Return -> clear buffer
	brne	carriagereturn_end
	ldi	buffer, RAM_Buffer
carriagereturn_end:
	rjmp	tim0_ret

save_char:
	;printable char (#32 .. #255): save to buffer
	cpi	buffer, RAM_BufferEnd
	brsh	tim0_ret
	push	ZL
	mov	ZL, buffer
	st	Z, u_buffer
	pop	ZL
	inc	buffer

tim0_ret:
	out	SREG,u_sr		;Restore status register
	reti

;===============================================================================

;Software UART:
uart_init:	
	ldi	u_tmp,u_R
	out	TCCR0,u_tmp		;Start timer and set clock source
	ldi	u_tmp,1<<INT0
	out	GIMSK,u_tmp		;Enable external interrupt 0
	ldi	u_tmp,1<<ISC01
	out	MCUCR,u_tmp		;On falling edges
	clr	flags			;Erase status-byte
	ret

;===============================================================================

start:
	;set stackpointer
	ldi	temp, LOW(RAMEND)
	out	SPL, temp

	;init ports
	ldi	temp, 0x03
	out	DDRA, temp
	ldi	temp, 0x00
	out	PORTA, temp
	
	ldi	temp, 0x00
	out	DDRB, temp
	ldi	temp, 0xFF
	out	PORTB, temp

	ldi	temp, 0x79
	out	DDRD, temp
	ldi	temp, 0x0E
	out	PORTD, temp

	;initialize software UART (appnote AVR304)
	rcall	uart_init

	;init registers
	ldi	YH, 0
	ldi	ZH, 0
	ldi	buffer, RAM_Buffer
	
	;init display RAM
	rcall	clear

	;enable interrupts (for software UART)
	sei

;===============================================================================

changed:
	;text changed, clear display and restart static/scroller
	rcall	clear
	andi	flags, ~(1<<RDR | 1<<Scroll)
	cpi	buffer, RAM_Buffer+17 ;scroll if >= 17 chars
	brlo	static_init
	ori	flags, 1<<Scroll
	rjmp	scroll_loop

static_init:
	;initialize static display
	mov	count, buffer
	subi	count, RAM_Buffer
	breq	static_loop ;empty buffer: display nothing
	ldi	ZL, RAM_Buffer
	ldi	YL, RAM_Disp
static_init_loop:
	ld	temp, Z+
	st	Y+, temp
	dec	count
	brne	static_init_loop
	rjmp	static_loop
	

scroll_loop:
	;restart scroller
	ldi	ZL, RAM_Buffer
	
text_loop:
	;scroll by 1 char
	ld	temp, Z
text_next:
	adiw	ZH:ZL, 1
	ldi	count, 15
	ldi	YL, RAM_Disp
ramshift:
	ldd	temp2, Y+1
	st	Y+, temp2
	dec	count
	brne	ramshift
	st	Y, temp

	;display current string
	ldi	count, 15 ;scroller speed: 1 count = ~10ms
static_loop:
	ldi	grid, 0
mux:
	;get char
	ldi	YL, RAM_Disp
	add	YL, grid
	ld	temp, Y+
	;get values for 14 segment display
	push	ZL
	push	ZH
	ldi	ZL, LOW(charset*2)
	ldi	ZH, HIGH(charset*2)
	add	ZL, temp
	brcc	PC+2
	inc	ZH
	add	ZL, temp
	brcc	PC+2
	inc	ZH
	lpm	disp1, Z+
	lpm	disp2, Z+
	pop	ZH
	pop	ZL
	;display char
	rcall	shift
	mov	temp, grid
	swap	temp
	andi	temp, 0x70
	ori	temp, 1<<E | 1<<RW | 1<<RS | 1<<Strobe
	out	PORTD, temp
	nop
	cbi	PORTD, Strobe
	;delay
	ldi	temp, 5
	ldi	temp2, 0
delay:
	dec	temp2
	brne	delay
	dec	temp
	brne	delay
	;next grid
	inc	grid
	cpi	grid, 16
	brlo	mux
	sbrc	flags, RDR ;did the string change?
	rjmp	changed
	dec	count
	brne	static_loop
	sbrs	flags, Scroll ;loop indefinitely if not scrolling
	rjmp	static_loop

	;loop end for scroller
	cp	ZL, buffer
	brlo	text_loop
	;add " +++ " to the end of the string
	cp	ZL, buffer
	brne	PC+3
	ldi	temp, ' '
	rjmp	text_next
	mov	temp, buffer
	subi	temp, -4
	cp	ZL, temp
	brne	PC+3
	ldi	temp, ' '
	rjmp	text_next
	inc	temp
	cp	ZL, temp
	brlo	PC+2
	rjmp	scroll_loop
	ldi	temp, '+'
	rjmp	text_next
	

;--------------------

shift:	;invert and shift out disp1[0..7], disp2[0..6] and grid[3]
	mov	temp, disp1
	ldi	temp2, 8
shift1:
	sbi	PORTA, Data
	lsr	temp
	brcc	PC+2
	cbi	PORTA, Data
	sbi	PORTA, Clock
	nop
	cbi	PORTA, Clock
	dec	temp2
	brne	shift1
	
	mov	temp, disp2
	ldi	temp2, 7
shift2:
	sbi	PORTA, Data
	lsr	temp
	brcc	PC+2
	cbi	PORTA, Data
	sbi	PORTA, Clock
	nop
	cbi	PORTA, Clock
	dec	temp2
	brne	shift2

	cbi	PORTA, Data
	sbrc	grid, 3
	sbi	PORTA, Data
	sbi	PORTA, Clock
	nop
	cbi	PORTA, Clock

	ret

;--------------------

clear:
	;clear display RAM
	ldi	count, 16
	ldi	YL, RAM_Disp
	ldi	temp, ' '
clear_loop:
	st	Y+, temp
	dec	count
	brne	clear_loop
	ret

;===============================================================================

charset:
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 ;0x00
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 ;0x10
	.dw 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000
	.dw 0x0000, 0x0709, 0x0220, 0x12ce, 0x12ed, 0x0c24, 0x24f9, 0x0200 ;0x20
	.dw 0x2400, 0x0900, 0x3fc0, 0x12c0, 0x0800, 0x00c0, 0x4000, 0x0c00
	.dw 0x0c3f, 0x0406, 0x088b, 0x048d, 0x00e6, 0x2069, 0x00fd, 0x0c01 ;0x30
	.dw 0x00ff, 0x00ef, 0x0009, 0x0801, 0x2400, 0x00c8, 0x0900, 0x1083
	.dw 0x08af, 0x00f7, 0x128f, 0x0039, 0x120f, 0x0079, 0x0071, 0x00bd ;0x40
	.dw 0x00f6, 0x1209, 0x001e, 0x2470, 0x0038, 0x0536, 0x2136, 0x003f
	.dw 0x00f3, 0x203f, 0x20f3, 0x00ed, 0x1201, 0x003e, 0x0c30, 0x2836 ;0x50
	.dw 0x2d00, 0x0d00, 0x0c09, 0x0039, 0x2100, 0x000f, 0x0001, 0x0008
	.dw 0x0100, 0x00f7, 0x128f, 0x0039, 0x120f, 0x0079, 0x0071, 0x00bd ;0x60
	.dw 0x00f6, 0x1209, 0x001e, 0x2470, 0x0038, 0x0536, 0x2136, 0x003f
	.dw 0x00f3, 0x203f, 0x20f3, 0x00ed, 0x1201, 0x003e, 0x0c30, 0x2836 ;0x70
	.dw 0x2d00, 0x0d00, 0x0c09, 0x2440, 0x1200, 0x0980, 0x0001, 0x2808
	